@@ -58,7 +58,7 @@ size_t appctx_raw_snd_buf(struct appctx *appctx, struct buffer *buf, size_t coun
 size_t appctx_snd_buf(struct stconn *sc, struct buffer *buf, size_t count, unsigned int flags);
 
 int appctx_fastfwd(struct stconn *sc, unsigned int count, unsigned int flags);
-ssize_t applet_append_line(void *ctx, struct ist v1, struct ist v2, size_t ofs, size_t len);
+ssize_t applet_append_line(void *ctx, struct ist v1, struct ist v2, size_t ofs, size_t len, char delim);
 static forceinline void applet_fl_set(struct appctx *appctx, uint on);
 static forceinline void applet_fl_clr(struct appctx *appctx, uint off);
 
@@ -116,7 +116,7 @@ static inline int appctx_init(struct appctx *appctx)
 	 * the appctx will be fully initialized. The session and the stream will
 	 * eventually be created. The affinity must be set now !
 	 */
-	BUG_ON(appctx->t->tid != -1 && appctx->t->tid != tid);
+	BUG_ON(appctx->t->tid != tid);
 	task_set_thread(appctx->t, tid);
 
 	if (appctx->applet->init)
@@ -282,16 +282,16 @@ static inline void applet_expect_data(struct appctx *appctx)
 	se_fl_clr(appctx->sedesc, SE_FL_EXP_NO_DATA);
 }
 
-/* writes chunk <chunk> into the input channel of the stream attached to this
- * appctx's endpoint, and marks the SC_FL_NEED_ROOM on a channel full error.
- * See ci_putchk() for the list of return codes.
- */
-static inline int applet_putchk(struct appctx *appctx, struct buffer *chunk)
+/* Should only be used via wrappers applet_putchk() / applet_putchk_stress(). */
+static inline int _applet_putchk(struct appctx *appctx, struct buffer *chunk,
+                                 int stress)
 {
 	int ret;
 
 	if (appctx->flags & APPCTX_FL_INOUT_BUFS) {
-		if (b_data(chunk) > b_room(&appctx->outbuf)) {
+		if (unlikely(stress) ?
+		    b_data(&appctx->outbuf) :
+		    b_data(chunk) > b_room(&appctx->outbuf)) {
 			applet_fl_set(appctx, APPCTX_FL_OUTBLK_FULL);
 			ret = -1;
 		}
@@ -303,8 +303,8 @@ static inline int applet_putchk(struct appctx *appctx, struct buffer *chunk)
 	else {
 		struct sedesc *se = appctx->sedesc;
 
-		ret = ci_putchk(sc_ic(se->sc), chunk);
-		if (ret < 0) {
+		if ((unlikely(stress) && ci_data(sc_ic(se->sc))) ||
+		    (ret = ci_putchk(sc_ic(se->sc), chunk)) < 0) {
 			/* XXX: Handle all errors as a lack of space because callers
 			 * don't handles other cases for now. So applets must be
 			 * careful to handles shutdown (-2) and invalid calls (-3) by
@@ -314,9 +314,25 @@ static inline int applet_putchk(struct appctx *appctx, struct buffer *chunk)
 			ret = -1;
 		}
 	}
+
 	return ret;
 }
 
+/* writes chunk <chunk> into the input channel of the stream attached to this
+ * appctx's endpoint, and marks the SC_FL_NEED_ROOM on a channel full error.
+ * See ci_putchk() for the list of return codes.
+ */
+static inline int applet_putchk(struct appctx *appctx, struct buffer *chunk)
+{
+	return _applet_putchk(appctx, chunk, 0);
+}
+
+/* Equivalent of applet_putchk() but with stress condition alternatives activated. */
+static inline int applet_putchk_stress(struct appctx *appctx, struct buffer *chunk)
+{
+	return _applet_putchk(appctx, chunk, 1);
+}
+
 /* writes <len> chars from <blk> into the input channel of the stream attached
  * to this appctx's endpoint, and marks the SC_FL_NEED_ROOM on a channel full
  * error. See ci_putblk() for the list of return codes.
