@@ -17,6 +17,7 @@
 #include <haproxy/quic_frame-t.h>
 #include <haproxy/quic_pacing-t.h>
 #include <haproxy/quic_stream-t.h>
+#include <haproxy/quic_utils-t.h>
 #include <haproxy/stconn-t.h>
 #include <haproxy/task-t.h>
 #include <haproxy/time-t.h>
@@ -27,16 +28,20 @@ enum qcs_type {
 	QCS_SRV_BIDI,
 	QCS_CLT_UNI,
 	QCS_SRV_UNI,
-
-	/* Must be the last one */
-	QCS_MAX_TYPES
 };
 
+enum qcc_app_st {
+	QCC_APP_ST_NULL,
+	QCC_APP_ST_INIT,
+	QCC_APP_ST_SHUT,
+} __attribute__((packed));
+
 struct qcc {
 	struct connection *conn;
 	uint64_t nb_sc; /* number of attached stream connectors */
 	uint64_t nb_hreq; /* number of in-progress http requests */
 	uint32_t flags; /* QC_CF_* */
+	enum qcc_app_st app_st; /* application layer state */
 	int glitches;   /* total number of glitches on this connection */
 
 	/* flow-control fields set by us enforced on our side. */
@@ -107,9 +112,6 @@ struct qcc {
 	void *ctx; /* Application layer context */
 };
 
-/* Maximum size of stream Rx buffer. */
-#define QC_S_RX_BUF_SZ   (global.tune.bufsize - NCB_RESERVED_SZ)
-
 /* QUIC stream states
  *
  * On initialization a stream is put on idle state. It is opened as soon as
@@ -132,6 +134,15 @@ enum qcs_state {
 	QC_SS_CLO,      /* closed */
 } __attribute__((packed));
 
+/* STREAM receive buffer. Can handle out-of-order storage.
+ * Can be used as a tree node to allocate multiple entries ordered by offsets.
+ */
+struct qc_stream_rxbuf {
+	struct eb64_node off_node; /* base offset of current buffer, node for QCS rx.bufs */
+	struct ncbuf ncb;          /* data storage with support for out of order offset */
+	uint64_t off_end;          /* first offset directly outside of current buffer */
+};
+
 struct qcs {
 	struct qcc *qcc;
 	struct sedesc *sd;
@@ -140,15 +151,17 @@ struct qcs {
 	void *ctx;           /* app-ops context */
 
 	struct {
-		uint64_t offset; /* absolute current base offset of ncbuf */
+		uint64_t offset; /* read offset */
 		uint64_t offset_max; /* maximum absolute offset received */
-		struct ncbuf ncbuf; /* receive buffer - can handle out-of-order offset frames */
+		struct eb_root bufs; /* receive buffers tree ordered by offset */
 		struct buffer app_buf; /* receive buffer used by stconn layer */
 		uint64_t msd; /* current max-stream-data limit to enforce */
-		uint64_t msd_init; /* initial max-stream-data */
+		uint64_t msd_base; /* max-stream-data previous to latest update */
+		struct bdata_ctr data; /* data utilization counter. Note that <tot> is now used for now as accounting may be difficult with ncbuf. */
 	} rx;
 	struct {
 		struct quic_fctl fc; /* stream flow control applied on sending */
+		struct quic_frame *msd_frm; /* MAX_STREAM_DATA frame prepared */
 	} tx;
 
 	struct eb64_node by_id;
@@ -195,7 +208,7 @@ struct qcc_app_ops {
 	ssize_t (*rcv_buf)(struct qcs *qcs, struct buffer *b, int fin);
 
 	/* Convert HTX to HTTP payload for sending. */
-	size_t (*snd_buf)(struct qcs *qcs, struct buffer *b, size_t count, char *fin);
+	size_t (*snd_buf)(struct qcs *qcs, struct buffer *b, size_t count);
 
 	/* Negotiate and commit fast-forward data from opposite MUX. */
 	size_t (*nego_ff)(struct qcs *qcs, size_t count);
@@ -223,7 +236,7 @@ struct qcc_app_ops {
 #define QC_CF_ERRL_DONE 0x00000002 /* local error properly handled, connection can be released */
 /* unused 0x00000004 */
 #define QC_CF_CONN_FULL 0x00000008 /* no stream buffers available on connection */
-#define QC_CF_APP_SHUT  0x00000010 /* Application layer shutdown done. */
+/* unused 0x00000010 */
 #define QC_CF_ERR_CONN  0x00000020 /* fatal error reported by transport layer */
 #define QC_CF_WAIT_HS   0x00000040 /* MUX init before QUIC handshake completed (0-RTT) */
 
@@ -240,9 +253,8 @@ static forceinline char *qcc_show_flags(char *buf, size_t len, const char *delim
 	_(QC_CF_ERRL,
 	_(QC_CF_ERRL_DONE,
 	_(QC_CF_CONN_FULL,
-	_(QC_CF_APP_SHUT,
 	_(QC_CF_ERR_CONN,
-	_(QC_CF_WAIT_HS))))));
+	_(QC_CF_WAIT_HS)))));
 	/* epilogue */
 	_(~0U);
 	return buf;
