@@ -46,7 +46,28 @@
 
 #ifdef USE_QUIC_OPENSSL_COMPAT
 #include <haproxy/quic_openssl_compat.h>
+#else
+#if defined(OSSL_FUNC_SSL_QUIC_TLS_CRYPTO_SEND)
+/* This macro is defined by the new OpenSSL 3.5.0 QUIC TLS API and it is not
+ * defined by quictls.
+ */
+#define HAVE_OPENSSL_QUIC
+#define SSL_set_quic_transport_params   SSL_set_quic_tls_transport_params
+#define SSL_set_quic_early_data_enabled SSL_set_quic_tls_early_data_enabled
+#define SSL_quic_read_level(arg) -1
+
+enum ssl_encryption_level_t {
+	ssl_encryption_initial = 0,
+	ssl_encryption_early_data,
+	ssl_encryption_handshake,
+	ssl_encryption_application
+};
+
+#else
+/* QUIC TLS API */
+#define HAVE_OPENSSL_QUICTLS
 #endif
+#endif /* USE_QUIC_OPENSSL_COMPAT */
 
 #if defined(OPENSSL_IS_AWSLC)
 #define OPENSSL_NO_DH
@@ -128,6 +149,14 @@
 #define HAVE_SSL_SET_SECURITY_LEVEL
 #endif
 
+#if ((defined(LIBRESSL_VERSION_NUMBER) && (LIBRESSL_VERSION_NUMBER >= 0x3030600L)) || (HA_OPENSSL_VERSION_NUMBER >= 0x10101000L) || defined(OPENSSL_IS_AWSLC)) && !defined(USE_OPENSSL_WOLFSSL)
+#define HAVE_JWS
+#endif
+
+#if (defined(HAVE_JWS) && defined(HAVE_ASN1_TIME_TO_TM))
+#define HAVE_ACME
+#endif
+
 #if !defined(HAVE_SSL_SET_SECURITY_LEVEL)
 /* define a nope function for set_security_level */
 #define SSL_CTX_set_security_level(ctx, level) ({})
@@ -149,6 +178,12 @@
 #define HASSL_DH_up_ref DH_up_ref
 #endif
 
+#if ((defined SSL_CTRL_SET_TLSEXT_STATUS_REQ_CB || AWSLC_API_VERSION >= 29) && (!defined(OPENSSL_NO_OCSP)))
+#define HAVE_SSL_OCSP
+#else
+typedef void OCSP_CERTID;
+#endif
+
 #if ((HA_OPENSSL_VERSION_NUMBER < 0x1000000fL) && !defined(X509_get_X509_PUBKEY))
 #define X509_get_X509_PUBKEY(x) ((x)->cert_info->key)
 #endif
@@ -186,19 +221,6 @@ static inline STACK_OF(X509) *X509_chain_up_ref(STACK_OF(X509) *chain)
 
 #endif
 
-#ifdef OPENSSL_IS_BORINGSSL
-/*
- * Functions missing in BoringSSL
- */
-
-static inline X509_CRL *X509_OBJECT_get0_X509_CRL(const X509_OBJECT *a)
-{
-    if (a == NULL || a->type != X509_LU_CRL) {
-        return NULL;
-    }
-    return a->data.crl;
-}
-#endif
 
 #if (HA_OPENSSL_VERSION_NUMBER < 0x1010000fL) && (!defined(LIBRESSL_VERSION_NUMBER) || LIBRESSL_VERSION_NUMBER < 0x2070000fL)
 /*
@@ -411,7 +433,7 @@ static inline unsigned long ERR_peek_error_func(const char **func)
 #define SSL_OP_CIPHER_SERVER_PREFERENCE 0
 #endif
 
-/* needs OpenSSL >= 0.9.7 and renegotation options on WolfSSL */
+/* needs OpenSSL >= 0.9.7 and renegotiation options on WolfSSL */
 #if !defined(SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION) || \
 	 (defined(USE_OPENSSL_WOLFSSL) && !defined(HAVE_SECURE_RENEGOTIATION) && !defined(HAVE_SERVER_RENEGOTIATION_INFO))
 #undef  SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION
