@@ -86,7 +86,7 @@ static inline size_t qc_frm_len(struct quic_frame *frm)
 	}
 	case QUIC_FT_CRYPTO: {
 		struct qf_crypto *f = &frm->crypto;
-		len += 1 + quic_int_getsize(f->offset_node.key) + quic_int_getsize(f->len) + f->len;
+		len += 1 + quic_int_getsize(f->offset) + quic_int_getsize(f->len) + f->len;
 		break;
 	}
 	case QUIC_FT_NEW_TOKEN: {
@@ -206,7 +206,7 @@ static inline struct quic_err quic_err_app(uint64_t code)
  *
  * Returns the allocated frame or NULL on failure.
  */
-static inline struct quic_frame *qc_frm_alloc(int type)
+static inline struct quic_frame *qc_frm_alloc(uint64_t type)
 {
 	struct quic_frame *frm = NULL;
 
@@ -277,5 +277,8 @@ static inline void qc_stream_frm_mv_fwd(struct quic_frame *frm, uint64_t data)
 	strm_frm->data = (unsigned char *)b_peek(&cf_buf, data);
 }
 
+size_t quic_strm_frm_fillbuf(size_t room, struct quic_frame *frm, size_t *split_size);
+struct quic_frame *quic_strm_frm_split(struct quic_frame *frm, uint64_t left);
+
 #endif /* USE_QUIC */
 #endif /* _HAPROXY_QUIC_FRAME_H */
