@@ -67,6 +67,10 @@ struct ckch_conf {
 	char *issuer;
 	char *sctl;
 	int ocsp_update_mode;
+	struct {
+		char *id;
+		char **domains;
+	} acme;
 };
 
 /*
@@ -83,6 +87,7 @@ struct ckch_store {
 	struct list ckch_inst; /* list of ckch_inst which uses this ckch_node */
 	struct list crtlist_entry; /* list of entries which use this store */
 	struct ckch_conf conf;
+	struct task *acme_task;
 	struct ebmb_node node;
 	char path[VAR_ARRAY];
 };
@@ -108,11 +113,6 @@ struct ckch_inst_link_ref {
        struct list list;
 };
 
-
-#define CKCH_INST_NO_DEFAULT          0   /* not the default instance */
-#define CKCH_INST_IMPL_DEFAULT        1   /* implicit default instance (declaration order) */
-#define CKCH_INST_EXPL_DEFAULT        2   /* explicit default instance (default-crt) */
-
 /*
  * This structure describe a ckch instance. An instance is generated for each
  * bind_conf.  The instance contains a linked list of the sni ctx which uses
@@ -125,7 +125,7 @@ struct ckch_inst {
 	struct crtlist_entry *crtlist_entry; /* pointer to the crtlist_entry used, or NULL */
 	struct server *server; /* pointer to the server if is_server_instance is set, NULL otherwise */
 	SSL_CTX *ctx; /* pointer to the SSL context used by this instance */
-	unsigned int is_default:2;      /* This instance is used as the default ctx for this bind_conf (1: implicit default, 2: explicit default) */
+	unsigned int is_default:1;      /* This instance is used as the default ctx for this bind_conf */
 	unsigned int is_server_instance:1; /* This instance is used by a backend server */
 	/* space for more flag there */
 	struct list sni_ctx; /* list of sni_ctx using this ckch_inst */
@@ -188,16 +188,16 @@ struct cert_exts {
 enum parse_type_t {
 	PARSE_TYPE_NONE = 0,
 	PARSE_TYPE_INT,
-	PARSE_TYPE_STR,         /* string which is strdup() */
-	PARSE_TYPE_ONOFF,       /* "on" or "off" keyword */
+	PARSE_TYPE_STR,            /* string which is strdup() */
+	PARSE_TYPE_ARRAY_SUBSTR,   /* string split by colons into an array of substring */
+	PARSE_TYPE_ONOFF,          /* "on" or "off" keyword */
 };
 
 struct ckch_conf_kws {
 	const char *name;
 	ssize_t offset;
 	enum parse_type_t type;
-	int (*func)(void *value, char *buf, struct ckch_data *d, int cli, char **err);
-	char **base; /* ptr to the base path */
+	int (*func)(void *value, char *buf, struct ckch_data *d, int cli, const char *filename, int linenum, char **err);
 };
 
 extern struct ckch_conf_kws ckch_conf_kws[];
