@@ -258,7 +258,6 @@ struct ssl_sock_ctx {
 	unsigned long error_code;     /* last error code of the error stack */
 	struct buffer early_buf;      /* buffer to store the early data received */
 	int sent_early_data;          /* Amount of early data we sent so far */
-	int can_send_early_data;      /* We did not start the handshake yet so we can send early data */
 
 #ifdef USE_QUIC
 	struct quic_conn *qc;
@@ -318,7 +317,9 @@ struct global_ssl {
 	} ocsp_update;
 #endif
 
-	unsigned int certificate_compression:1; /* allow to explicitely disable certificate compression */
+#ifdef HAVE_ACME
+	int acme_scheduler;
+#endif
 };
 
 /* The order here matters for picking a default context,
@@ -334,6 +335,8 @@ struct ssl_counters {
 	long long sess;
 	long long reused_sess;
 	long long failed_handshake;
+	long long ocsp_staple;
+	long long failed_ocsp_staple;
 };
 
 #endif /* USE_OPENSSL */
