@@ -88,7 +88,7 @@ static void *memprof_realloc_initial_handler(void *ptr, size_t size);
 static char *memprof_strdup_initial_handler(const char *s);
 static void  memprof_free_initial_handler(void *ptr);
 
-/* these ones are optional but may be used by some dependecies */
+/* these ones are optional but may be used by some dependencies */
 static char *memprof_strndup_initial_handler(const char *s, size_t n);
 static void *memprof_valloc_initial_handler(size_t sz);
 static void *memprof_pvalloc_initial_handler(size_t sz);
@@ -105,7 +105,7 @@ static void *(*memprof_realloc_handler)(void *ptr, size_t size)   = memprof_real
 static char *(*memprof_strdup_handler)(const char *s)             = memprof_strdup_initial_handler;
 static void  (*memprof_free_handler)(void *ptr)                   = memprof_free_initial_handler;
 
-/* these ones are optional but may be used by some dependecies */
+/* these ones are optional but may be used by some dependencies */
 static char *(*memprof_strndup_handler)(const char *s, size_t n)                 = memprof_strndup_initial_handler;
 static void *(*memprof_valloc_handler)(size_t sz)                                = memprof_valloc_initial_handler;
 static void *(*memprof_pvalloc_handler)(size_t sz)                               = memprof_pvalloc_initial_handler;
@@ -1327,7 +1327,7 @@ static int cli_io_handler_show_tasks(struct appctx *appctx)
 			t = eb32_entry(rqnode, struct task, rq);
 			entry = sched_activity_entry(tmp_activity, t->process, NULL);
 			if (t->wake_date) {
-				lat = (uint32_t)now_ns - t->wake_date;
+				lat = now_ns - t->wake_date;
 				if ((int64_t)lat > 0)
 					entry->lat_time += lat;
 			}
@@ -1344,7 +1344,7 @@ static int cli_io_handler_show_tasks(struct appctx *appctx)
 			t = eb32_entry(rqnode, struct task, rq);
 			entry = sched_activity_entry(tmp_activity, t->process, NULL);
 			if (t->wake_date) {
-				lat = (uint32_t)now_ns - t->wake_date;
+				lat = now_ns - t->wake_date;
 				if ((int64_t)lat > 0)
 					entry->lat_time += lat;
 			}
@@ -1357,7 +1357,7 @@ static int cli_io_handler_show_tasks(struct appctx *appctx)
 			t = (const struct task *)tl;
 			entry = sched_activity_entry(tmp_activity, t->process, NULL);
 			if (!TASK_IS_TASKLET(t) && t->wake_date) {
-				lat = (uint32_t)now_ns - t->wake_date;
+				lat = now_ns - t->wake_date;
 				if ((int64_t)lat > 0)
 					entry->lat_time += lat;
 			}
@@ -1370,7 +1370,7 @@ static int cli_io_handler_show_tasks(struct appctx *appctx)
 				t = (const struct task *)tl;
 				entry = sched_activity_entry(tmp_activity, t->process, NULL);
 				if (!TASK_IS_TASKLET(t) && t->wake_date) {
-					lat = (uint32_t)now_ns - t->wake_date;
+					lat = now_ns - t->wake_date;
 					if ((int64_t)lat > 0)
 						entry->lat_time += lat;
 				}
