@@ -463,9 +463,15 @@ void conn_init(struct connection *conn, void *target)
 	conn->send_proxy_ofs = 0;
 	conn->handle.fd = DEAD_FD_MAGIC;
 	conn->err_code = CO_ER_NONE;
+	conn->term_evts_log = 0;
 	conn->target = target;
 	conn->destroy_cb = NULL;
 	conn->proxy_netns = NULL;
+	MT_LIST_INIT(&conn->toremove_list);
+	if (conn_is_back(conn))
+		LIST_INIT(&conn->sess_el);
+	else
+		LIST_INIT(&conn->stopping_list);
 	LIST_INIT(&conn->tlv_list);
 	conn->subs = NULL;
 	conn->src = NULL;
@@ -474,7 +480,6 @@ void conn_init(struct connection *conn, void *target)
 	conn->xprt = NULL;
 	conn->reverse.target = NULL;
 	conn->reverse.name = BUF_NULL;
-	conn->sni_hash = 0;
 }
 
 /* Initialize members used for backend connections.
@@ -483,10 +488,6 @@ void conn_init(struct connection *conn, void *target)
  */
 static int conn_backend_init(struct connection *conn)
 {
-	LIST_INIT(&conn->idle_list);
-	LIST_INIT(&conn->sess_el);
-	MT_LIST_INIT(&conn->toremove_list);
-
 	if (!sockaddr_alloc(&conn->dst, 0, 0))
 		return 1;
 
@@ -497,12 +498,6 @@ static int conn_backend_init(struct connection *conn)
 	return 0;
 }
 
-/* Initialize members used only on frontend connections. */
-static void conn_frontend_init(struct connection *conn)
-{
-	LIST_INIT(&conn->stopping_list);
-}
-
 /* Release connection elements reserved for backend side usage. It also takes
  * care to detach it if linked to a session or a server instance.
  *
@@ -515,10 +510,15 @@ static void conn_backend_deinit(struct connection *conn)
 	if (LIST_INLIST(&conn->sess_el))
 		session_unown_conn(conn->owner, conn);
 
-	/* If the connection is not private, it is accounted by the server. */
-	if (!(conn->flags & CO_FL_PRIVATE)) {
-		if (obj_type(conn->target) == OBJ_TYPE_SERVER)
-			srv_release_conn(__objt_server(conn->target), conn);
+	if (obj_type(conn->target) == OBJ_TYPE_SERVER) {
+		struct server *srv = __objt_server(conn->target);
+
+		/* If the connection is not private, it is accounted by the server. */
+		if (!(conn->flags & CO_FL_PRIVATE)) {
+			srv_release_conn(srv, conn);
+		}
+		if (srv->flags & SRV_F_STRICT_MAXCONN)
+			_HA_ATOMIC_DEC(&srv->curr_total_conns);
 	}
 
 	/* Make sure the connection is not left in the idle connection tree */
@@ -528,18 +528,6 @@ static void conn_backend_deinit(struct connection *conn)
 	pool_free(pool_head_conn_hash_node, conn->hash_node);
 	conn->hash_node = NULL;
 
-	/* Remove from BE purge list. Necessary if conn already scheduled for
-	 * purge but finally freed before by another code path.
-	 */
-	MT_LIST_DELETE(&conn->toremove_list);
-}
-
-/* Ensure <conn> frontend connection is removed from its lists. This must be
- * performed before freeing or reversing a connection.
- */
-static void conn_frontend_deinit(struct connection *conn)
-{
-	LIST_DEL_INIT(&conn->stopping_list);
 }
 
 /* Tries to allocate a new connection and initialized its main fields. The
@@ -565,9 +553,6 @@ struct connection *conn_new(void *target)
 			return NULL;
 		}
 	}
-	else {
-		conn_frontend_init(conn);
-	}
 
 	return conn;
 }
@@ -579,8 +564,14 @@ void conn_free(struct connection *conn)
 
 	if (conn_is_back(conn))
 		conn_backend_deinit(conn);
-	else
-		conn_frontend_deinit(conn);
+
+	/* Remove the conn from toremove_list.
+	 *
+	 * This is needed to prevent a double-free in case the connection was
+	 * already scheduled from cleaning but is freed before via another
+	 * call.
+	 */
+	MT_LIST_DELETE(&conn->toremove_list);
 
 	sockaddr_free(&conn->src);
 	sockaddr_free(&conn->dst);
@@ -1395,11 +1386,14 @@ int conn_recv_proxy(struct connection *conn, int flag)
 	goto fail;
 
  recv_abort:
+	conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_truncated_shutr);
 	conn->err_code = CO_ER_PRX_ABORT;
 	conn->flags |= CO_FL_SOCK_RD_SH | CO_FL_SOCK_WR_SH;
 	goto fail;
 
  fail:
+	if (!(conn->flags & CO_FL_SOCK_RD_SH))
+		conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_truncated_rcv_err);
 	conn->flags |= CO_FL_ERROR;
 	return 0;
 }
@@ -1493,6 +1487,7 @@ int conn_send_proxy(struct connection *conn, unsigned int flag)
 
  out_error:
 	/* Write error on the file descriptor */
+	conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_snd_err);
 	conn->flags |= CO_FL_ERROR;
 	return 0;
 
@@ -1695,11 +1690,14 @@ int conn_recv_netscaler_cip(struct connection *conn, int flag)
 	goto fail;
 
  recv_abort:
+	conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_truncated_shutr);
 	conn->err_code = CO_ER_CIP_ABORT;
 	conn->flags |= CO_FL_SOCK_RD_SH | CO_FL_SOCK_WR_SH;
 	goto fail;
 
  fail:
+	if (!(conn->flags & CO_FL_SOCK_RD_SH))
+		conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_truncated_rcv_err);
 	conn->flags |= CO_FL_ERROR;
 	return 0;
 }
@@ -1773,6 +1771,7 @@ int conn_send_socks4_proxy_request(struct connection *conn)
 
  out_error:
 	/* Write error on the file descriptor */
+	conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_snd_err);
 	conn->flags |= CO_FL_ERROR;
 	if (conn->err_code == CO_ER_NONE) {
 		conn->err_code = CO_ER_SOCKS4_SEND;
@@ -1893,6 +1892,7 @@ int conn_recv_socks4_proxy_response(struct connection *conn)
 	return 0;
 
  recv_abort:
+	conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_truncated_shutr);
 	if (conn->err_code == CO_ER_NONE) {
 		conn->err_code = CO_ER_SOCKS4_ABORT;
 	}
@@ -1900,6 +1900,8 @@ int conn_recv_socks4_proxy_response(struct connection *conn)
 	goto fail;
 
  fail:
+	if (!(conn->flags & CO_FL_SOCK_RD_SH))
+		conn_report_term_evt(conn, tevt_loc_hs, hs_tevt_type_truncated_rcv_err);
 	conn->flags |= CO_FL_ERROR;
 	return 0;
 }
@@ -2720,7 +2722,7 @@ static struct sample_fetch_kw_list sample_fetch_keywords = {ILH, {
 	{ "bc_glitches", smp_fetch_fc_glitches, 0, NULL, SMP_T_SINT, SMP_USE_L4SRV },
 	{ "bc_http_major", smp_fetch_fc_http_major, 0, NULL, SMP_T_SINT, SMP_USE_L4SRV },
 	{ "bc_nb_streams", smp_fetch_fc_nb_streams, 0, NULL, SMP_T_SINT, SMP_USE_L5SRV },
-	{ "bc_settings_streams_limit", smp_fetch_fc_streams_limit, 0, NULL, SMP_T_SINT, SMP_USE_L5SRV },
+	{ "bc_setting_streams_limit", smp_fetch_fc_streams_limit, 0, NULL, SMP_T_SINT, SMP_USE_L5SRV },
 	{ "fc_err", smp_fetch_fc_err, 0, NULL, SMP_T_SINT, SMP_USE_L4CLI },
 	{ "fc_err_name", smp_fetch_fc_err_str, 0, NULL, SMP_T_STR, SMP_USE_L4CLI },
 	{ "fc_err_str", smp_fetch_fc_err_str, 0, NULL, SMP_T_STR, SMP_USE_L4CLI },
@@ -2929,8 +2931,6 @@ int conn_reverse(struct connection *conn)
 		struct server *srv = objt_server(conn->reverse.target);
 		BUG_ON(!srv);
 
-		conn_frontend_deinit(conn);
-
 		if (conn_backend_init(conn))
 			return 1;
 
@@ -2954,6 +2954,7 @@ int conn_reverse(struct connection *conn)
 		srv_use_conn(srv, conn);
 
 		/* Free the session after detaching the connection from it. */
+		session_unown_conn(sess, conn);
 		sess->origin = NULL;
 		session_free(sess);
 		conn_set_owner(conn, NULL, NULL);
@@ -2966,7 +2967,6 @@ int conn_reverse(struct connection *conn)
 
 		conn_backend_deinit(conn);
 
-		conn_frontend_init(conn);
 		conn->target = &l->obj_type;
 		conn->flags |= CO_FL_ACT_REVERSING;
 		task_wakeup(l->rx.rhttp.task, TASK_WOKEN_RES);
