@@ -39,6 +39,7 @@
 #include <haproxy/quic_enc.h>
 #include <haproxy/quic_fctl.h>
 #include <haproxy/quic_frame.h>
+#include <haproxy/quic_utils.h>
 #include <haproxy/stats-t.h>
 #include <haproxy/tools.h>
 #include <haproxy/trace.h>
@@ -148,10 +149,9 @@ struct h3c {
 
 DECLARE_STATIC_POOL(pool_head_h3c, "h3c", sizeof(struct h3c));
 
-#define H3_SF_UNI_INIT     0x00000001  /* stream type not parsed for unidirectional stream */
-#define H3_SF_UNI_NO_H3    0x00000002  /* unidirectional stream does not carry H3 frames */
-#define H3_SF_HAVE_CLEN    0x00000004  /* content-length header is present */
-#define H3_SF_SENT_INTERIM 0x00000008  /* last response sent is 1xx interim. Used on FE side only. */
+#define H3_SF_UNI_INIT  0x00000001  /* stream type not parsed for unidirectional stream */
+#define H3_SF_UNI_NO_H3 0x00000002  /* unidirectional stream does not carry H3 frames */
+#define H3_SF_HAVE_CLEN 0x00000004  /* content-length header is present */
 
 struct h3s {
 	struct h3c *h3c;
@@ -530,6 +530,7 @@ static ssize_t h3_headers_to_htx(struct qcs *qcs, const struct buffer *buf,
 	unsigned int flags = HTX_SL_F_NONE;
 	struct ist meth = IST_NULL, path = IST_NULL;
 	struct ist scheme = IST_NULL, authority = IST_NULL;
+	struct ist uri;
 	struct ist v;
 	int hdr_idx, ret;
 	int cookie = -1, last_cookie = -1, i;
@@ -735,11 +736,6 @@ static ssize_t h3_headers_to_htx(struct qcs *qcs, const struct buffer *buf,
 			goto out;
 		}
 	}
-	else {
-		h3s->err = H3_ERR_REQUEST_REJECTED;
-		len = -1;
-		goto out;
-	}
 
 	flags |= HTX_SL_F_VER_11;
 	flags |= HTX_SL_F_XFER_LEN;
@@ -778,9 +774,41 @@ static ssize_t h3_headers_to_htx(struct qcs *qcs, const struct buffer *buf,
 		goto out;
 	}
 
+	if (!istlen(scheme)) {
+		/* No scheme (CONNECT), use :authority only. */
+		uri = authority;
+		flags |= HTX_SL_F_HAS_AUTHORITY;
+	}
+	else if (isttest(authority)) {
+		/* Use absolute URI form as :authority is present. */
+		uri = ist2bin(trash.area, scheme);
+		istcat(&uri, ist("://"), trash.size);
+		istcat(&uri, authority, trash.size);
+		if (!isteq(path, ist("*")))
+			istcat(&uri, path, trash.size);
+
+		flags |= HTX_SL_F_HAS_AUTHORITY;
+		if (flags & (HTX_SL_F_SCHM_HTTP|HTX_SL_F_SCHM_HTTPS)) {
+			/* we don't know if it was originally an absolute or a
+			 * relative request because newer versions of HTTP use
+			 * the absolute URI format by default, which we call
+			 * the normalized URI format internally. This is the
+			 * strongly recommended way of sending a request for
+			 * a regular client, so we cannot distinguish this
+			 * from a request intended for a proxy. For other
+			 * schemes however there is no doubt.
+			 */
+			flags |= HTX_SL_F_NORMALIZED_URI;
+		}
+	}
+	else {
+		/* Use origin URI form. */
+		uri = path;
+	}
+
 	/* Ensure that final URI does not contains LWS nor CTL characters. */
-	for (i = 0; i < path.len; i++) {
-		unsigned char c = istptr(path)[i];
+	for (i = 0; i < uri.len; i++) {
+		unsigned char c = istptr(uri)[i];
 		if (HTTP_IS_LWS(c) || HTTP_IS_CTL(c)) {
 			TRACE_ERROR("invalid character in path", H3_EV_RX_FRAME|H3_EV_RX_HDR, qcs->qcc->conn, qcs);
 			h3s->err = H3_ERR_MESSAGE_ERROR;
@@ -790,7 +818,7 @@ static ssize_t h3_headers_to_htx(struct qcs *qcs, const struct buffer *buf,
 		}
 	}
 
-	sl = htx_add_stline(htx, HTX_BLK_REQ_SL, flags, meth, path, ist("HTTP/3.0"));
+	sl = htx_add_stline(htx, HTX_BLK_REQ_SL, flags, meth, uri, ist("HTTP/3.0"));
 	if (!sl) {
 		len = -1;
 		goto out;
@@ -823,7 +851,7 @@ static ssize_t h3_headers_to_htx(struct qcs *qcs, const struct buffer *buf,
 
 		for (i = 0; i < list[hdr_idx].n.len; ++i) {
 			const char c = list[hdr_idx].n.ptr[i];
-			if ((uint8_t)(c - 'A') <= 'Z' - 'A' || !HTTP_IS_TOKEN(c)) {
+			if ((uint8_t)(c - 'A') < 'Z' - 'A' || !HTTP_IS_TOKEN(c)) {
 				TRACE_ERROR("invalid characters in field name", H3_EV_RX_FRAME|H3_EV_RX_HDR, qcs->qcc->conn, qcs);
 				h3s->err = H3_ERR_MESSAGE_ERROR;
 				qcc_report_glitch(h3c->qcc, 1);
@@ -974,7 +1002,7 @@ static ssize_t h3_headers_to_htx(struct qcs *qcs, const struct buffer *buf,
 		}
 	}
 
-	/* Check the number of blocks agains "tune.http.maxhdr" value before adding EOH block */
+	/* Check the number of blocks against "tune.http.maxhdr" value before adding EOH block */
 	if (htx_nbblks(htx) > global.tune.max_http_hdr) {
 		len = -1;
 		goto out;
@@ -1110,7 +1138,7 @@ static ssize_t h3_trailers_to_htx(struct qcs *qcs, const struct buffer *buf,
 
 		for (i = 0; i < list[hdr_idx].n.len; ++i) {
 			const char c = list[hdr_idx].n.ptr[i];
-			if ((uint8_t)(c - 'A') <= 'Z' - 'A' || !HTTP_IS_TOKEN(c)) {
+			if ((uint8_t)(c - 'A') < 'Z' - 'A' || !HTTP_IS_TOKEN(c)) {
 				TRACE_ERROR("invalid characters in field name", H3_EV_RX_FRAME|H3_EV_RX_HDR, qcs->qcc->conn, qcs);
 				h3s->err = H3_ERR_MESSAGE_ERROR;
 				qcc_report_glitch(h3c->qcc, 1);
@@ -1172,7 +1200,7 @@ static ssize_t h3_trailers_to_htx(struct qcs *qcs, const struct buffer *buf,
 		++hdr_idx;
 	}
 
-	/* Check the number of blocks agains "tune.http.maxhdr" value before adding EOT block */
+	/* Check the number of blocks against "tune.http.maxhdr" value before adding EOT block */
 	if (htx_nbblks(htx) > global.tune.max_http_hdr) {
 		len = -1;
 		goto out;
@@ -1408,21 +1436,8 @@ static ssize_t h3_rcv_buf(struct qcs *qcs, struct buffer *b, int fin)
 	}
 
 	if (!b_data(b) && fin && quic_stream_is_bidi(qcs->id)) {
-		struct buffer *appbuf;
-		struct htx *htx;
-		int eom;
-
 		TRACE_PROTO("received FIN without data", H3_EV_RX_FRAME, qcs->qcc->conn, qcs);
-		if (!(appbuf = qcc_get_stream_rxbuf(qcs))) {
-			TRACE_ERROR("data buffer alloc failure", H3_EV_RX_FRAME, qcs->qcc->conn, qcs);
-			qcc_set_error(qcs->qcc, H3_ERR_INTERNAL_ERROR, 1);
-			goto err;
-		}
-
-		htx = htx_from_buf(appbuf);
-		eom = htx_set_eom(htx);
-		htx_to_buf(htx, appbuf);
-		if (!eom) {
+		if (qcs_http_handle_standalone_fin(qcs)) {
 			TRACE_ERROR("cannot set EOM", H3_EV_RX_FRAME, qcs->qcc->conn, qcs);
 			qcc_set_error(qcs->qcc, H3_ERR_INTERNAL_ERROR, 1);
 			goto err;
@@ -1479,7 +1494,7 @@ static ssize_t h3_rcv_buf(struct qcs *qcs, struct buffer *b, int fin)
 			 * SETTINGS_MAX_FIELD_SECTION_SIZE parameter to prevent
 			 * excessive decompressed size.
 			 */
-			if (flen > QC_S_RX_BUF_SZ) {
+			if (flen > qmux_stream_rx_bufsz()) {
 				TRACE_ERROR("received a too big frame", H3_EV_RX_FRAME, qcs->qcc->conn, qcs);
 				qcc_set_error(qcs->qcc, H3_ERR_EXCESSIVE_LOAD, 1);
 				qcc_report_glitch(qcs->qcc, 1);
@@ -1552,11 +1567,13 @@ static ssize_t h3_rcv_buf(struct qcs *qcs, struct buffer *b, int fin)
 	if (h3s->err) {
 		qcc_abort_stream_read(qcs);
 		qcc_reset_stream(qcs, h3s->err);
-		return b_data(b);
+		total = b_data(b);
+		goto done;
 	}
 	else if (h3c->err) {
 		qcc_set_error(qcs->qcc, h3c->err, 1);
-		return b_data(b);
+		total = b_data(b);
+		goto done;
 	}
 	else if (unlikely(ret < 0)) {
 		qcc_set_error(qcs->qcc, H3_ERR_INTERNAL_ERROR, 1);
@@ -1576,7 +1593,7 @@ static ssize_t h3_rcv_buf(struct qcs *qcs, struct buffer *b, int fin)
 	return -1;
 }
 
-/* Function used to emit stream data from <qcs> control uni-stream.
+/* Emit SETTINGS frame on <qcs> control uni-stream.
  *
  * On success return the number of sent bytes. A negative code is used on
  * error.
@@ -1619,11 +1636,6 @@ static int h3_control_send(struct qcs *qcs, void *ctx)
 		b_quic_enc_int(&pos, h3_settings_max_field_section_size, 0);
 	}
 
-	if (qfctl_sblocked(&qcs->tx.fc) || qfctl_sblocked(&qcs->qcc->tx.fc)) {
-		TRACE_ERROR("not enough initial credit for control stream", H3_EV_TX_FRAME|H3_EV_TX_SETTINGS, qcs->qcc->conn, qcs);
-		goto err;
-	}
-
 	if (!(res = qcc_get_stream_txbuf(qcs, &err, 0))) {
 		/* Only memory failure can cause buf alloc error for control stream due to qcs_send_metadata() usage. */
 		TRACE_ERROR("cannot allocate Tx buffer", H3_EV_TX_FRAME|H3_EV_TX_SETTINGS, qcs->qcc->conn, qcs);
@@ -1676,7 +1688,6 @@ static int h3_encode_header(struct buffer *buf,
 
 static int h3_resp_headers_send(struct qcs *qcs, struct htx *htx)
 {
-	struct h3s *h3s = qcs->ctx;
 	int err;
 	struct buffer outbuf;
 	struct buffer headers_buf = BUF_NULL;
@@ -1708,20 +1719,8 @@ static int h3_resp_headers_send(struct qcs *qcs, struct htx *htx)
 			/* start-line -> HEADERS h3 frame */
 			BUG_ON(sl);
 			sl = htx_get_blk_ptr(htx, blk);
+			/* TODO should be on h3 layer */
 			status = sl->info.res.status;
-			if (status < 100 || status > 999) {
-				TRACE_ERROR("invalid response status code", H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
-				goto err;
-			}
-			else if (status >= 100 && status < 200) {
-				TRACE_USER("handling interim HTX response", H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
-				BUG_ON(conn_is_back(qcs->qcc->conn)); /* H3_SF_SENT_INTERIM is FE side only. */
-				h3s->flags |= H3_SF_SENT_INTERIM;
-			}
-			else {
-				TRACE_USER("handling final HTX response", H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
-				h3s->flags &= ~H3_SF_SENT_INTERIM;
-			}
 		}
 		else if (type == HTX_BLK_HDR) {
 			if (unlikely(hdr >= sizeof(list) / sizeof(list[0]) - 1)) {
@@ -1739,9 +1738,6 @@ static int h3_resp_headers_send(struct qcs *qcs, struct htx *htx)
 		}
 	}
 
-	/* Current function expects HTX start-line to be present. This also
-	 * ensures <status> conformance has been checked prior to encoding it.
-	 */
 	BUG_ON(!sl);
 
 	list[hdr].n = ist("");
@@ -1759,26 +1755,20 @@ static int h3_resp_headers_send(struct qcs *qcs, struct htx *htx)
 		goto end;
 	}
 
-	/* Reserve space for frame type + length as a max varint size. */
-	if (unlikely(b_contig_space(res) < 5)) {
-		/* Most of the times, h3_resp_headers_send() is only called one
-		 * time per stream, so buffer will be empty. However, this
-		 * assumption is invalid when handling interim responses, so
-		 * it's important to check out buffer remaining space.
-		 */
-		goto err_full;
-	}
+	/* Buffer allocated just now : must be enough for frame type + length as a max varint size */
+	BUG_ON(b_room(res) < 5);
 
 	b_reset(&outbuf);
 	outbuf = b_make(b_tail(res), b_contig_space(res), 0, 0);
 	/* Start the headers after frame type + length */
-	headers_buf = b_make(b_tail(res) + 5, b_contig_space(res) - 5, 0, 0);
+	headers_buf = b_make(b_head(res) + 5, b_size(res) - 5, 0, 0);
 
 	TRACE_DATA("encoding HEADERS frame", H3_EV_TX_FRAME|H3_EV_TX_HDR,
 	           qcs->qcc->conn, qcs);
 	if (qpack_encode_field_section_line(&headers_buf))
 		goto err_full;
 	if (qpack_encode_int_status(&headers_buf, status)) {
+		/* TODO handle invalid status code VS no buf space left */
 		TRACE_ERROR("error during status code encoding", H3_EV_TX_FRAME|H3_EV_TX_HDR, qcs->qcc->conn, qcs);
 		goto err_full;
 	}
@@ -1818,23 +1808,10 @@ static int h3_resp_headers_send(struct qcs *qcs, struct htx *htx)
 	/* Now that all headers are encoded, we are certain that res buffer is
 	 * big enough
 	 */
-
-	/* Encode H3 HEADERS frame type + frame length. If buffer is empty,
-	 * length is encoded on min varint size and buf head is realigned.
-	 * However, if it still contains a previously encoded interim response
-	 * not yet emitted, buf head must not be adjusted and frame length is
-	 * encoded as a 4-bytes varint so that responses are contiguous.
-	 */
-	if (!b_data(res)) {
-		frame_length_size = quic_int_getsize(b_data(&headers_buf));
-		res->head += 4 - frame_length_size;
-		b_putchr(res, 0x01); /* h3 HEADERS frame type */
-		b_quic_enc_int(res, b_data(&headers_buf), 0);
-	}
-	else {
-		b_putchr(res, 0x01); /* h3 HEADERS frame type */
-		b_quic_enc_int(res, b_data(&headers_buf), 4);
-	}
+	frame_length_size = quic_int_getsize(b_data(&headers_buf));
+	res->head += 4 - frame_length_size;
+	b_putchr(res, 0x01); /* h3 HEADERS frame type */
+	b_quic_enc_int(res, b_data(&headers_buf), 0);
 	b_add(res, b_data(&headers_buf));
 
 	ret = 0;
@@ -1852,20 +1829,6 @@ static int h3_resp_headers_send(struct qcs *qcs, struct htx *htx)
 	return ret;
 
  err_full:
-	if (b_data(res)) {
-		/* Output buffer already contains data : this may happens after
-		 * HTTP interim response encoding. Try to release buffer to be
-		 * able to encode newer interim or final response.
-		 */
-		if (qcc_release_stream_txbuf(qcs)) {
-			TRACE_DEVEL("cannot release buf", H3_EV_TX_FRAME|H3_EV_TX_HDR, qcs->qcc->conn, qcs);
-			return 0;
-		}
-
-		TRACE_DEVEL("retry after buf release", H3_EV_TX_FRAME|H3_EV_TX_HDR, qcs->qcc->conn, qcs);
-		goto retry;
-	}
-
 	if (smallbuf) {
 		TRACE_DEVEL("retry with a full buffer", H3_EV_TX_FRAME|H3_EV_TX_HDR, qcs->qcc->conn, qcs);
 		smallbuf = 0;
@@ -1902,6 +1865,11 @@ static int h3_resp_trailers_send(struct qcs *qcs, struct htx *htx)
 	TRACE_ENTER(H3_EV_TX_FRAME|H3_EV_TX_HDR, qcs->qcc->conn, qcs);
 
 	hdr = 0;
+
+	/* Skip the trailers because the corresponding conf option was set */
+	if (qcs->qcc->proxy->options & PR_O_HTTP_DROP_RES_TRLS)
+		goto skip_trailers;
+
 	for (blk = htx_get_head_blk(htx); blk; blk = htx_get_next_blk(htx, blk)) {
 		type = htx_get_blk_type(blk);
 
@@ -1926,6 +1894,7 @@ static int h3_resp_trailers_send(struct qcs *qcs, struct htx *htx)
 		}
 	}
 
+ skip_trailers:
 	if (!hdr) {
 		/* No headers encoded here so no need to generate a H3 HEADERS
 		 * frame. Mux will send an empty QUIC STREAM frame with FIN.
@@ -2178,24 +2147,19 @@ static int h3_resp_data_send(struct qcs *qcs, struct htx *htx,
 	return -1;
 }
 
-static size_t h3_snd_buf(struct qcs *qcs, struct buffer *buf, size_t count, char *fin)
+static size_t h3_snd_buf(struct qcs *qcs, struct buffer *buf, size_t count)
 {
-	struct h3s *h3s = qcs->ctx;
 	size_t total = 0;
 	enum htx_blk_type btype;
 	struct htx *htx;
 	struct htx_blk *blk;
 	uint32_t bsize;
 	int32_t idx;
-	char eom;
 	int ret = 0;
 
 	TRACE_ENTER(H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
 
-	*fin = 0;
 	htx = htx_from_buf(buf);
-	/* EOM is saved here, useful if 0-copy is performed with HTX buf. */
-	eom = htx->flags & HTX_FL_EOM;
 
 	while (count && !htx_is_empty(htx) && qcc_stream_can_send(qcs) && ret >= 0) {
 		idx = htx_get_head(htx);
@@ -2208,7 +2172,7 @@ static size_t h3_snd_buf(struct qcs *qcs, struct buffer *buf, size_t count, char
 
 		switch (btype) {
 		case HTX_BLK_RES_SL:
-			/* start-line -> HEADERS h3 frame (FE side) */
+			/* start-line -> HEADERS h3 frame */
 			ret = h3_resp_headers_send(qcs, htx);
 			if (ret > 0) {
 				total += ret;
@@ -2289,10 +2253,6 @@ static size_t h3_snd_buf(struct qcs *qcs, struct buffer *buf, size_t count, char
 #endif
 
  out:
-	if (eom && htx_is_empty(htx) && !(h3s->flags & H3_SF_SENT_INTERIM)) {
-		TRACE_USER("transcoding last HTX message", H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
-		*fin = 1;
-	}
 	htx_to_buf(htx, buf);
 
 	TRACE_LEAVE(H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
@@ -2357,9 +2317,14 @@ static size_t h3_nego_ff(struct qcs *qcs, size_t count)
 	return ret;
 }
 
+/* Finalize encoding of a HTTP/3 data frame after zero-copy for <qcs> stream.
+ * No frame is encoded if <qcs> iobuf indicates that no data were transferred.
+ *
+ * Return the payload size of the H3 data frame or 0 if no frame encoded.
+ */
 static size_t h3_done_ff(struct qcs *qcs)
 {
-	size_t total = qcs->sd->iobuf.data;
+	size_t total = 0;
 	TRACE_ENTER(H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
 
 	h3_debug_printf(stderr, "%s\n", __func__);
@@ -2367,15 +2332,14 @@ static size_t h3_done_ff(struct qcs *qcs)
 	if (qcs->sd->iobuf.data) {
 		TRACE_DATA("encoding DATA frame (fast forward)",
 		           H3_EV_TX_FRAME|H3_EV_TX_DATA, qcs->qcc->conn, qcs);
+
 		b_sub(qcs->sd->iobuf.buf, qcs->sd->iobuf.data);
 		b_putchr(qcs->sd->iobuf.buf, 0x00); /* h3 frame type = DATA */
 		b_quic_enc_int(qcs->sd->iobuf.buf, qcs->sd->iobuf.data, QUIC_VARINT_MAX_SIZE); /* h3 frame length */
 		b_add(qcs->sd->iobuf.buf, qcs->sd->iobuf.data);
-	}
 
-	qcs->sd->iobuf.buf = NULL;
-	qcs->sd->iobuf.offset = 0;
-	qcs->sd->iobuf.data = 0;
+		total = qcs->sd->iobuf.offset + qcs->sd->iobuf.data;
+	}
 
 	TRACE_LEAVE(H3_EV_STRM_SEND, qcs->qcc->conn, qcs);
 	return total;
@@ -2580,44 +2544,61 @@ static int h3_init(struct qcc *qcc)
 	return 0;
 }
 
-/* Initialize H3 control stream and prepare SETTINGS emission.
+/* Open control stream for <ctx> HTTP/3 connection and schedule a SETTINGS
+ * frame emission on it.
  *
- * Returns 0 on success else non-zero.
+ * Returns 0 on success. If a transient error was encountered, a positive value
+ * is returned, finalize operation should be recalled later. A negative value is
+ * used for a fatal error, in this case the connection should be closed.
  */
 static int h3_finalize(void *ctx)
 {
 	struct h3c *h3c = ctx;
 	struct qcc *qcc = h3c->qcc;
-	struct qcs *qcs;
+	struct qcs *qcs = h3c->ctrl_strm;
 
 	TRACE_ENTER(H3_EV_H3C_NEW, qcc->conn);
 
-	/* RFC 9114 6.2. Unidirectional Streams
-	 *
-	 * Each endpoint needs to create at least one unidirectional stream for
-	 * the HTTP control stream. QPACK requires two additional unidirectional
-	 * streams, and other extensions might require further streams.
-	 * Therefore, the transport parameters sent by both clients and servers
-	 * MUST allow the peer to create at least three unidirectional streams.
-	 */
-	if (qcc_fctl_avail_streams(qcc, 0) < 3) {
-		TRACE_ERROR("peer flow-control limit does not allow control stream creation", H3_EV_H3C_NEW, qcc->conn);
-		qcc_set_error(qcc, H3_ERR_GENERAL_PROTOCOL_ERROR, 1);
-		qcc_report_glitch(qcc, 1);
-		goto err;
-	}
-
-	qcs = qcc_init_stream_local(qcc, 0);
 	if (!qcs) {
-		/* Error must be set by qcc_init_stream_local(). */
-		BUG_ON(!(qcc->flags & QC_CF_ERRL));
-		TRACE_ERROR("cannot init control stream", H3_EV_H3C_NEW, qcc->conn);
-		goto err;
+		/* RFC 9114 6.2. Unidirectional Streams
+		 *
+		 * Each endpoint needs to create at least one unidirectional stream for
+		 * the HTTP control stream. QPACK requires two additional unidirectional
+		 * streams, and other extensions might require further streams.
+		 * Therefore, the transport parameters sent by both clients and servers
+		 * MUST allow the peer to create at least three unidirectional streams.
+		 */
+		if (qcc_fctl_avail_streams(qcc, 0) < 3) {
+			TRACE_ERROR("peer flow-control limit does not allow control stream creation", H3_EV_H3C_NEW, qcc->conn);
+			qcc_set_error(qcc, H3_ERR_GENERAL_PROTOCOL_ERROR, 1);
+			qcc_report_glitch(qcc, 1);
+			goto err;
+		}
+
+		qcs = qcc_init_stream_local(qcc, 0);
+		if (!qcs) {
+			/* Error must be set by qcc_init_stream_local(). */
+			BUG_ON(!(qcc->flags & QC_CF_ERRL));
+			TRACE_ERROR("cannot init control stream", H3_EV_H3C_NEW, qcc->conn);
+			goto err;
+		}
+
+		qcs_send_metadata(qcs);
+		h3c->ctrl_strm = qcs;
 	}
 
-	qcs_send_metadata(qcs);
-	h3c->ctrl_strm = qcs;
+	if (qfctl_sblocked(&qcs->tx.fc) || qfctl_sblocked(&qcs->qcc->tx.fc)) {
+		TRACE_STATE("not enough initial credit for control stream", H3_EV_TX_FRAME|H3_EV_TX_SETTINGS, qcs->qcc->conn, qcs);
+		goto again;
+	}
 
+	/* RFC 9114 7.2.4.2. Initialization
+	 *
+	 * Endpoints MUST NOT require any data to be
+	 * received from the peer prior to sending the SETTINGS frame;
+	 * settings MUST be sent as soon as the transport is ready to
+	 * send data.
+	 */
 	if (h3_control_send(qcs, h3c) < 0) {
 		qcc_set_error(qcc, H3_ERR_INTERNAL_ERROR, 1);
 		goto err;
@@ -2626,9 +2607,13 @@ static int h3_finalize(void *ctx)
 	TRACE_LEAVE(H3_EV_H3C_NEW, qcc->conn);
 	return 0;
 
+ again:
+	TRACE_DEVEL("leaving on transient state", H3_EV_H3C_NEW, qcc->conn);
+	return 1;
+
  err:
 	TRACE_DEVEL("leaving on error", H3_EV_H3C_NEW, qcc->conn);
-	return 1;
+	return -1;
 }
 
 /* Send a HTTP/3 GOAWAY followed by a CONNECTION_CLOSE_APP. */
