@@ -388,8 +388,8 @@ int http_replace_req_uri(struct htx *htx, const struct ist uri)
 	if (!http_replace_stline(htx, meth, uri, vsn))
 		goto fail;
 
-	sl = http_get_stline(htx);
-	ALREADY_CHECKED(sl); /* the stline exists because http_replace_stline() succeeded */
+	/* the stline exists because http_replace_stline() succeeded */
+	sl = ASSUME_NONNULL(http_get_stline(htx));
 
 	sl->flags &= ~HTX_SL_F_NORMALIZED_URI;
 	if (!http_update_host(htx, sl, uri))
