@@ -7,6 +7,7 @@
 #include <haproxy/mux_quic.h>
 #include <haproxy/quic_conn-t.h>
 #include <haproxy/quic_frame-t.h>
+#include <haproxy/quic_utils.h>
 
 /* trace source and events */
 static void qmux_trace(enum trace_level level, uint64_t mask,
@@ -130,6 +131,16 @@ static void qmux_trace_fill_ctx(struct trace_ctx *ctx, const struct trace_source
 /* register qmux traces */
 INITCALL1(STG_REGISTER, trace_register_source, TRACE_SOURCE);
 
+static char *qcc_app_st_to_str(const enum qcc_app_st st)
+{
+	switch (st) {
+	case QCC_APP_ST_NULL: return "NULL";
+	case QCC_APP_ST_INIT: return "INIT";
+	case QCC_APP_ST_SHUT: return "SHUT";
+	default:              return "";
+	}
+}
+
 void qmux_dump_qcc_info(struct buffer *msg, const struct qcc *qcc)
 {
 	const struct quic_conn *qc = qcc->conn->handle.qc;
@@ -137,7 +148,9 @@ void qmux_dump_qcc_info(struct buffer *msg, const struct qcc *qcc)
 	chunk_appendf(msg, " qcc=%p(F)", qcc);
 	if (qcc->conn->handle.qc)
 		chunk_appendf(msg, " qc=%p", qcc->conn->handle.qc);
-	chunk_appendf(msg, " .sc=%llu .hreq=%llu .flg=0x%04x", (ullong)qcc->nb_sc, (ullong)qcc->nb_hreq, qcc->flags);
+	chunk_appendf(msg, " .st=%s .sc=%llu .hreq=%llu .flg=0x%04x",
+	              qcc_app_st_to_str(qcc->app_st), (ullong)qcc->nb_sc,
+	              (ullong)qcc->nb_hreq, qcc->flags);
 
 	chunk_appendf(msg, " .tx=%llu %llu/%llu bwnd=%llu/%llu",
 	              (ullong)qcc->tx.fc.off_soft, (ullong)qcc->tx.fc.off_real, (ullong)qcc->tx.fc.limit,
@@ -149,11 +162,16 @@ void qmux_dump_qcs_info(struct buffer *msg, const struct qcs *qcs)
 	chunk_appendf(msg, " qcs=%p .id=%llu .st=%s .flg=0x%04x", qcs, (ullong)qcs->id,
 	              qcs_st_to_str(qcs->st), qcs->flags);
 
-	chunk_appendf(msg, " .rx=%llu/%llu", (ullong)qcs->rx.offset_max, (ullong)qcs->rx.msd);
+	chunk_appendf(msg, " .rx=%llu/%llu rxb=%u(%u)",
+	              (ullong)qcs->rx.offset_max, (ullong)qcs->rx.msd,
+	              qcs->rx.data.bcnt, qcs->rx.data.bmax);
 	chunk_appendf(msg, " .tx=%llu %llu/%llu", (ullong)qcs->tx.fc.off_soft,
 	                                          (ullong)qcs->tx.fc.off_real,
 	                                          (ullong)qcs->tx.fc.limit);
 
+	if (qcs->stream)
+		bdata_ctr_print(msg, &qcs->stream->data, " buf=");
+
 	chunk_appendf(msg, " .ti=%u/%u/%u",
 	              tot_time_read(&qcs->timer.base),
 	              tot_time_read(&qcs->timer.buf),
