@@ -60,8 +60,9 @@
 #include <haproxy/quic_token.h>
 #include <haproxy/quic_tp.h>
 #include <haproxy/quic_trace.h>
+#include <haproxy/quic_tune.h>
 #include <haproxy/quic_tx.h>
-#include <haproxy/cbuf.h>
+#include <haproxy/quic_utils.h>
 #include <haproxy/proto_quic.h>
 #include <haproxy/quic_tls.h>
 #include <haproxy/ssl_sock.h>
@@ -519,28 +520,18 @@ int quic_build_post_handshake_frames(struct quic_conn *qc)
 			goto err;
 		}
 
-		conn_id = quic_cid_alloc(qc);
+		conn_id = new_quic_cid(qc->cids, qc, NULL, NULL);
 		if (!conn_id) {
 			qc_frm_free(qc, &frm);
 			TRACE_ERROR("CID allocation error", QUIC_EV_CONN_IO_CB, qc);
 			goto err;
 		}
 
-		if (quic_cid_generate(conn_id)) {
-			qc_frm_free(qc, &frm);
-			pool_free(pool_head_quic_connection_id, conn_id);
-			TRACE_ERROR("error on CID generation", QUIC_EV_CONN_IO_CB, qc);
-			goto err;
-		}
+		/* TODO To prevent CID tree locking, all CIDs created here
+		 * could be allocated at the same time as the first one.
+		 */
+		_quic_cid_insert(conn_id);
 
-		if (quic_cid_insert(conn_id, NULL)) {
-			TRACE_STATE("collision during CID generation on post-handshake", QUIC_EV_CONN_IO_CB, qc);
-			pool_free(pool_head_quic_connection_id, conn_id);
-			qc_frm_free(qc, &frm);
-			continue;
-		}
-
-		quic_cid_register_seq_num(conn_id);
 		quic_connection_id_to_frm_cpy(frm, conn_id);
 		LIST_APPEND(&frm_list, &frm->list);
 	}
@@ -634,7 +625,8 @@ struct task *quic_conn_app_io_cb(struct task *t, void *context, unsigned int sta
 
  out:
 	if ((qc->flags & QUIC_FL_CONN_CLOSING) && qc->mux_state != QC_MUX_READY) {
-		quic_conn_release(qc);
+		if (quic_conn_release(qc))
+			t = NULL;
 		qc = NULL;
 	}
 
@@ -761,7 +753,6 @@ static struct quic_conn_closed *qc_new_cc_conn(struct quic_conn *qc)
 	cc_qc->idle_timer_task->context = cc_qc;
 	cc_qc->idle_expire = qc->idle_expire;
 
-	cc_qc->cntrs = qc->cntrs;
 	cc_qc->conn = qc->conn;
 	qc->conn = NULL;
 
@@ -840,9 +831,9 @@ struct task *quic_conn_io_cb(struct task *t, void *context, unsigned int state)
 		/* Note: if no token for address validation was received
 		 * for a 0RTT connection, some 0RTT packet could still be
 		 * waiting for HP removal AFTER the successful handshake completion.
-		 * Indeed a successful handshake completion implicitely valids
+		 * Indeed a successful handshake completion implicitly valids
 		 * the peer address. In this case, one wants to process
-		 * these ORTT packets AFTER the succesful handshake completion.
+		 * these ORTT packets AFTER the successful handshake completion.
 		 *
 		 * On the contrary, when a token for address validation was received,
 		 * release 0RTT packets still waiting for HP removal. These
@@ -913,7 +904,8 @@ struct task *quic_conn_io_cb(struct task *t, void *context, unsigned int state)
 	}
 
 	if ((qc->flags & QUIC_FL_CONN_CLOSING) && qc->mux_state != QC_MUX_READY) {
-		quic_conn_release(qc);
+		if (quic_conn_release(qc))
+			t = NULL;
 		qc = NULL;
 	}
 
@@ -1038,7 +1030,6 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
                               struct sockaddr_storage *peer_addr,
                               int server, int token, void *owner)
 {
-	int i;
 	struct quic_conn *qc = NULL;
 	struct listener *l = server ? owner : NULL;
 	struct proxy *prx = l ? l->bind_conf->frontend : NULL;
@@ -1117,6 +1108,9 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
 	quic_tls_ku_reset(&qc->ku.nxt_rx);
 	quic_tls_ku_reset(&qc->ku.nxt_tx);
 
+#ifdef HAVE_OPENSSL_QUIC
+	qc->prot_level = OSSL_RECORD_PROTECTION_LEVEL_NONE;
+#endif
 	/* Encryption levels */
 	qc->iel = qc->eel = qc->hel = qc->ael = NULL;
 	LIST_INIT(&qc->qel_list);
@@ -1127,14 +1121,6 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
 	/* Required to safely call quic_conn_prx_cntrs_update() from quic_conn_release(). */
 	qc->prx_counters = NULL;
 
-	qc->cids = pool_alloc(pool_head_quic_cids);
-	if (!qc->cids) {
-		TRACE_ERROR("Could not allocate a new CID tree", QUIC_EV_CONN_INIT, qc);
-		goto err;
-	}
-	*qc->cids = EB_ROOT;
-	qc->next_cid_seq_num = 0;
-
 	/* QUIC Server (or listener). */
 	if (server) {
 		cc_algo = l->bind_conf->quic_cc_algo;
@@ -1183,8 +1169,17 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
 		goto err;
 	}
 
-	if (l && HA_ATOMIC_LOAD(&l->rx.quic_mode) == QUIC_SOCK_MODE_CONN &&
-	    (global.tune.options & GTUNE_QUIC_SOCK_PER_CONN) &&
+	qc->cids = pool_alloc(pool_head_quic_cids);
+	if (!qc->cids) {
+		TRACE_ERROR("Could not allocate a new CID tree", QUIC_EV_CONN_INIT, qc);
+		goto err;
+	}
+	*qc->cids = EB_ROOT;
+
+	conn_id->qc = qc;
+
+	if (HA_ATOMIC_LOAD(&l->rx.quic_mode) == QUIC_SOCK_MODE_CONN &&
+	    (quic_tune.options & QUIC_TUNE_SOCK_PER_CONN) &&
 	    is_addr(local_addr)) {
 		TRACE_USER("Allocate a socket for QUIC connection", QUIC_EV_CONN_INIT, qc);
 		qc_alloc_fd(qc, local_addr, peer_addr);
@@ -1220,8 +1215,7 @@ struct quic_conn *qc_new_conn(const struct quic_version *qv, int ipv4,
 	qc->bytes.rx = 0;
 	memset(&qc->rx.params, 0, sizeof(qc->rx.params));
 	qc->rx.buf = b_make(qc->rx.buf.area, QUIC_CONN_RX_BUFSZ, 0, 0);
-	for (i = 0; i < QCS_MAX_TYPES; i++)
-		qc->rx.strms[i].nb_streams = 0;
+	qc->rx.stream_max_uni = qc->rx.stream_max_bidi = 0;
 
 	qc->nb_pkt_for_cc = 1;
 	qc->nb_pkt_since_cc = 0;
@@ -1400,12 +1394,15 @@ static inline void quic_conn_prx_cntrs_update(struct quic_conn *qc)
  *
  * This function must only be called by the thread responsible of the quic_conn
  * tasklet.
+ *
+ * Returns 1 if the tasklet was released, 0 otherwise
  */
-void quic_conn_release(struct quic_conn *qc)
+int quic_conn_release(struct quic_conn *qc)
 {
 	struct eb64_node *node;
 	struct quic_rx_packet *pkt, *pktback;
 	struct quic_conn_closed *cc_qc;
+	int ret = 0;
 
 	TRACE_ENTER(QUIC_EV_CONN_CLOSE, qc);
 
@@ -1433,6 +1430,7 @@ void quic_conn_release(struct quic_conn *qc)
 		qc->cids = NULL;
 		pool_free(pool_head_quic_cc_buf, qc->tx.cc_buf_area);
 		qc->tx.cc_buf_area = NULL;
+		ret = 1;
 	}
 
 	if (qc_test_fd(qc))
@@ -1450,6 +1448,12 @@ void quic_conn_release(struct quic_conn *qc)
 		HA_ATOMIC_DEC(&qc->li->rx.quic_curr_accept);
 	}
 
+	/* Substract last congestion window from global memory counter. */
+	if (qc->path) {
+		cshared_add(&quic_mem_diff, -qc->path->cwnd);
+		qc->path->cwnd = 0;
+	}
+
 	/* free remaining stream descriptors */
 	node = eb64_first(&qc->streams_by_id);
 	while (node) {
@@ -1529,6 +1533,7 @@ void quic_conn_release(struct quic_conn *qc)
 	TRACE_PROTO("QUIC conn. freed", QUIC_EV_CONN_FREED, qc);
  leave:
 	TRACE_LEAVE(QUIC_EV_CONN_CLOSE, qc);
+	return ret;
 }
 
 /* Initialize the timer task of <qc> QUIC connection.
@@ -1562,13 +1567,9 @@ void qc_idle_timer_do_rearm(struct quic_conn *qc, int arm_ack)
 {
 	unsigned int expire;
 
-	TRACE_ENTER(QUIC_EV_CONN_IDLE_TIMER, qc);
-
 	/* It is possible the idle timer task has been already released. */
-	if (!qc->idle_timer_task) {
-		TRACE_PROTO("idle timer already released", QUIC_EV_CONN_IDLE_TIMER, qc);
-		goto leave;
-	}
+	if (!qc->idle_timer_task)
+		return;
 
 	if (qc->flags & (QUIC_FL_CONN_CLOSING|QUIC_FL_CONN_DRAINING)) {
 		/* RFC 9000 10.2. Immediate Close
@@ -1597,8 +1598,6 @@ void qc_idle_timer_do_rearm(struct quic_conn *qc, int arm_ack)
 		expire = QUIC_MAX(3 * quic_pto(qc), qc->max_idle_timeout);
 	}
 
-	TRACE_PRINTF(TRACE_LEVEL_PROTO, QUIC_EV_CONN_IDLE_TIMER, qc, 0, 0, 0,
-	             "quic_pto=%u expire=%u", quic_pto(qc), expire);
 	qc->idle_expire = tick_add(now_ms, MS_TO_TICKS(expire));
 	/* Note that the ACK timer is not armed during the handshake. So,
 	 * the handshake expiration date is taken into an account only
@@ -1620,9 +1619,6 @@ void qc_idle_timer_do_rearm(struct quic_conn *qc, int arm_ack)
 		task_queue(qc->idle_timer_task);
 		TRACE_PROTO("idle timer armed", QUIC_EV_CONN_IDLE_TIMER, qc);
 	}
-
- leave:
-	TRACE_LEAVE(QUIC_EV_CONN_IDLE_TIMER, qc);
 }
 
 /* Rearm the idle timer or ack timer for <qc> QUIC connection depending on <read>
@@ -1878,7 +1874,7 @@ int qc_bind_tid_prep(struct quic_conn *qc, uint new_tid)
 	}
 
 	/* Reinit IO tasklet. */
-	if (qc->wait_event.tasklet->state & TASK_IN_LIST)
+	if (qc->wait_event.tasklet->state & TASK_QUEUED)
 		qc->flags |= QUIC_FL_CONN_IO_TO_REQUEUE;
 	tasklet_kill(qc->wait_event.tasklet);
 	/* In most cases quic_conn_app_io_cb is used but for 0-RTT quic_conn_io_cb can be still activated. */
@@ -1949,7 +1945,7 @@ void qc_bind_tid_commit(struct quic_conn *qc, struct listener *new_li)
 	conn_id = eb64_entry(node, struct quic_connection_id, seq_num);
 
 	/* Rebinding is considered done when CID points to the new
-	 * thread. quic-conn instance cannot be derefence after it.
+	 * thread. quic-conn instance cannot be dereferenced after it.
 	 */
 	HA_ATOMIC_STORE(&conn_id->tid, new_tid);
 	qc = NULL;
