@@ -16,7 +16,7 @@
 
 #include <haproxy/h3.h>
 #include <haproxy/list.h>
-#include <haproxy/ncbmbuf.h>
+#include <haproxy/ncbuf.h>
 #include <haproxy/proto_quic.h>
 #include <haproxy/quic_ack.h>
 #include <haproxy/quic_cc_drs.h>
@@ -31,6 +31,7 @@
 #include <haproxy/quic_token.h>
 #include <haproxy/quic_trace.h>
 #include <haproxy/quic_tx.h>
+#include <haproxy/quic_utils.h>
 #include <haproxy/ssl_sock.h>
 #include <haproxy/trace.h>
 
@@ -205,7 +206,7 @@ static int qc_pkt_decrypt(struct quic_conn *qc, struct quic_enc_level *qel,
 /* Handle <frm> frame whose packet it is attached to has just been acknowledged. The memory allocated
  * for this frame will be at least released in every cases.
  *
- * Returns 1 on sucess else 0.
+ * Returns 1 on success else 0.
  */
 static int qc_handle_newly_acked_frm(struct quic_conn *qc, struct quic_frame *frm)
 {
@@ -314,7 +315,7 @@ static void qc_newly_acked_pkts(struct quic_conn *qc, struct eb_root *pkts,
 
 /* Handle <newly_acked_pkts> list of newly acknowledged TX packets.
  *
- * Returns 1 on sucess else 0.
+ * Returns 1 on success else 0.
  */
 static int qc_handle_newly_acked_pkts(struct quic_conn *qc,
                                       unsigned int *pkt_flags, struct list *newly_acked_pkts)
@@ -661,19 +662,19 @@ static enum quic_rx_ret_frm qc_handle_crypto_frm(struct quic_conn *qc,
 	enum quic_rx_ret_frm ret = QUIC_RX_RET_FRM_DONE;
 	/* XXX TO DO: <cfdebug> is used only for the traces. */
 	struct quic_rx_crypto_frm cfdebug = {
-		.offset_node.key = crypto_frm->offset_node.key,
+		.offset_node.key = crypto_frm->offset,
 		.len = crypto_frm->len,
 	};
 	struct quic_cstream *cstream = qel->cstream;
-	struct ncbmbuf *ncbuf = &qel->cstream->rx.ncbuf;
+	struct ncbuf *ncbuf = &qel->cstream->rx.ncbuf;
 	uint64_t off_rel;
 
 	TRACE_ENTER(QUIC_EV_CONN_PRSHPKT, qc);
 
-	if (unlikely(crypto_frm->offset_node.key < cstream->rx.offset)) {
+	if (unlikely(crypto_frm->offset < cstream->rx.offset)) {
 		size_t diff;
 
-		if (crypto_frm->offset_node.key + crypto_frm->len <= cstream->rx.offset) {
+		if (crypto_frm->offset + crypto_frm->len <= cstream->rx.offset) {
 			/* Nothing to do */
 			TRACE_PROTO("Already received CRYPTO data",
 			            QUIC_EV_CONN_RXPKT, qc, pkt, &cfdebug);
@@ -684,19 +685,19 @@ static enum quic_rx_ret_frm qc_handle_crypto_frm(struct quic_conn *qc,
 		TRACE_PROTO("Partially already received CRYPTO data",
 		            QUIC_EV_CONN_RXPKT, qc, pkt, &cfdebug);
 
-		diff = cstream->rx.offset - crypto_frm->offset_node.key;
+		diff = cstream->rx.offset - crypto_frm->offset;
 		crypto_frm->len -= diff;
 		crypto_frm->data += diff;
-		crypto_frm->offset_node.key = cstream->rx.offset;
+		crypto_frm->offset = cstream->rx.offset;
 	}
 
-	if (!quic_get_ncbuf(ncbuf) || ncbmb_is_null(ncbuf)) {
+	if (!quic_get_ncbuf(ncbuf) || ncb_is_null(ncbuf)) {
 		TRACE_ERROR("CRYPTO ncbuf allocation failed", QUIC_EV_CONN_PRSHPKT, qc);
 		goto err;
 	}
 
 	/* crypto_frm->offset > cstream-trx.offset */
-	off_rel = crypto_frm->offset_node.key - cstream->rx.offset;
+	off_rel = crypto_frm->offset - cstream->rx.offset;
 
 	/* RFC 9000 7.5. Cryptographic Message Buffering
 	 *
@@ -706,18 +707,31 @@ static enum quic_rx_ret_frm qc_handle_crypto_frm(struct quic_conn *qc,
 	 * handshake. If an endpoint does not expand its buffer, it MUST close
 	 * the connection with a CRYPTO_BUFFER_EXCEEDED error code.
 	 */
-	if (off_rel + crypto_frm->len > ncbmb_size(ncbuf)) {
+	if (off_rel + crypto_frm->len > ncb_size(ncbuf)) {
 		TRACE_ERROR("CRYPTO frame too large", QUIC_EV_CONN_PRSHPKT, qc);
 		quic_set_connection_close(qc, quic_err_transport(QC_ERR_CRYPTO_BUFFER_EXCEEDED));
 		goto err;
 	}
 
-	ncb_ret = ncbmb_add(ncbuf, off_rel, (const char *)crypto_frm->data,
-	                    crypto_frm->len, NCB_ADD_OVERWRT);
-	BUG_ON(ncb_ret != NCB_RET_OK);
+	ncb_ret = ncb_add(ncbuf, off_rel, (const char *)crypto_frm->data,
+	                  crypto_frm->len, NCB_ADD_COMPARE);
+	if (ncb_ret != NCB_RET_OK) {
+		if (ncb_ret == NCB_RET_DATA_REJ) {
+			TRACE_ERROR("overlapping data rejected", QUIC_EV_CONN_PRSHPKT, qc);
+			quic_set_connection_close(qc, quic_err_transport(QC_ERR_PROTOCOL_VIOLATION));
+			qc_notify_err(qc);
+			goto err;
+		}
+		else if (ncb_ret == NCB_RET_GAP_SIZE) {
+			TRACE_DATA("cannot bufferize frame due to gap size limit",
+			           QUIC_EV_CONN_PRSHPKT, qc);
+			ret = QUIC_RX_RET_FRM_AGAIN;
+			goto done;
+		}
+	}
 
 	/* Reschedule with TASK_HEAVY if CRYPTO data ready for decoding. */
-	if (ncbmb_data(ncbuf, 0)) {
+	if (ncb_data(ncbuf, 0)) {
 		HA_ATOMIC_OR(&qc->wait_event.tasklet->state, TASK_HEAVY);
 		tasklet_wakeup(qc->wait_event.tasklet);
 	}
@@ -806,12 +820,13 @@ static inline unsigned int quic_ack_delay_ms(struct qf_ack *ack_frm,
 static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
                              struct quic_enc_level *qel)
 {
-	struct eb_root cf_frms_tree = EB_ROOT;
-	struct eb64_node *node;
-	struct quic_frame *frm = NULL;
+	struct list retry_frms = LIST_HEAD_INIT(retry_frms);
+	struct quic_frame *frm = NULL, *frm_tmp;
 	const unsigned char *pos, *end;
 	enum quic_rx_ret_frm ret;
-	int fast_retrans = 0, cf_err = 0;
+	int fast_retrans = 0;
+	/* parsing may be rerun multiple times, but no more than <iter>. */
+	int iter = 3, parsing_stage = 0;
 
 	TRACE_ENTER(QUIC_EV_CONN_PRSHPKT, qc);
 	/* Skip the AAD */
@@ -889,8 +904,36 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 			break;
 		}
 		case QUIC_FT_CRYPTO:
-			eb64_insert(&cf_frms_tree, &frm->crypto.offset_node);
-			frm = NULL;
+			ret = qc_handle_crypto_frm(qc, &frm->crypto, pkt, qel);
+			switch (ret) {
+			case QUIC_RX_RET_FRM_FATAL:
+				goto err;
+
+			case QUIC_RX_RET_FRM_AGAIN:
+				if (parsing_stage == 0) {
+					TRACE_STATE("parsing stage set to 1 (AGAIN encountered)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				/* Save frame in temp list to reparse it later. A new instance must be used for next packet frames. */
+				LIST_APPEND(&retry_frms, &frm->list);
+				frm = NULL;
+				break;
+
+			case QUIC_RX_RET_FRM_DUP:
+				if (qc_is_listener(qc) && qel == qc->iel &&
+				    !(qc->flags & QUIC_FL_CONN_HANDSHAKE_SPEED_UP)) {
+					fast_retrans = 1;
+				}
+				break;
+
+			case QUIC_RX_RET_FRM_DONE:
+				if (parsing_stage == 1) {
+					TRACE_STATE("parsing stage set to 2 (DONE after AGAIN)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				break;
+			}
+
 			break;
 		case QUIC_FT_NEW_TOKEN:
 			if (qc_is_listener(qc)) {
@@ -912,12 +955,13 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 		case QUIC_FT_STREAM_8 ... QUIC_FT_STREAM_F:
 		{
 			struct qf_stream *strm_frm = &frm->stream;
-			unsigned nb_streams = qc->rx.strms[qcs_id_type(strm_frm->id)].nb_streams;
 			const char fin = frm->type & QUIC_STREAM_FRAME_TYPE_FIN_BIT;
+			const uint64_t max = quic_stream_is_uni(strm_frm->id) ?
+			  qc->rx.stream_max_uni : qc->rx.stream_max_bidi;
 
 			/* The upper layer may not be allocated. */
 			if (qc->mux_state != QC_MUX_READY) {
-				if ((strm_frm->id >> QCS_ID_TYPE_SHIFT) < nb_streams) {
+				if (strm_frm->id < max) {
 					TRACE_DATA("Already closed stream", QUIC_EV_CONN_PRSHPKT, qc);
 				}
 				else {
@@ -982,7 +1026,6 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 		case QUIC_FT_RETIRE_CONNECTION_ID:
 		{
 			struct quic_connection_id *conn_id = NULL;
-			int retry_rand_cid = 3; /* Number of random retries on CID collision. */
 
 			if (!qc_handle_retire_connection_id_frm(qc, frm, &pkt->dcid, &conn_id))
 				goto err;
@@ -995,42 +1038,14 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 			pool_free(pool_head_quic_connection_id, conn_id);
 			TRACE_PROTO("CID retired", QUIC_EV_CONN_PSTRM, qc);
 
-			conn_id = quic_cid_alloc(qc);
+			conn_id = new_quic_cid(qc->cids, qc, NULL, NULL);
 			if (!conn_id) {
 				TRACE_ERROR("CID allocation error", QUIC_EV_CONN_IO_CB, qc);
-				quic_set_connection_close(qc, quic_err_transport(QC_ERR_INTERNAL_ERROR));
-				qc_notify_err(qc);
-				goto err;
 			}
-
-			while (retry_rand_cid--) {
-				if (quic_cid_generate(conn_id)) {
-					TRACE_ERROR("error on CID generation", QUIC_EV_CONN_PSTRM, qc);
-					quic_set_connection_close(qc, quic_err_transport(QC_ERR_INTERNAL_ERROR));
-					pool_free(pool_head_quic_connection_id, conn_id);
-					qc_notify_err(qc);
-					goto err;
-				}
-
-				if (quic_cid_insert(conn_id, NULL) == 0)
-					break;
+			else {
+				_quic_cid_insert(conn_id);
+				qc_build_new_connection_id_frm(qc, conn_id);
 			}
-
-			if (retry_rand_cid < 0) {
-				/* Multiple collisions during CID random gen.
-				 * Prefer to close the connection in error
-				 * immediately.
-				 */
-				TRACE_ERROR("CID pool exhausted", QUIC_EV_CONN_IO_CB, qc);
-				quic_set_connection_close(qc, quic_err_transport(QC_ERR_INTERNAL_ERROR));
-				pool_free(pool_head_quic_connection_id, conn_id);
-				qc_notify_err(qc);
-				goto err;
-			}
-
-			quic_cid_register_seq_num(conn_id);
-
-			qc_build_new_connection_id_frm(qc, conn_id);
 			break;
 		}
 		case QUIC_FT_PATH_CHALLENGE:
@@ -1084,45 +1099,57 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 	if (frm)
 		qc_frm_free(qc, &frm);
 
-	node = eb64_first(&cf_frms_tree);
-	while (node) {
-		frm = eb64_entry(node, struct quic_frame, crypto.offset_node);
-		/* only CRYPTO frames may be reparsed for now */
-		BUG_ON(frm->type != QUIC_FT_CRYPTO);
-		node = eb64_next(node);
-		ret = qc_handle_crypto_frm(qc, &frm->crypto, pkt, qel);
-		switch (ret) {
-		case QUIC_RX_RET_FRM_FATAL:
-			/* avoid freeing without eb_delete() */
-			frm = NULL;
+	while (!LIST_ISEMPTY(&retry_frms)) {
+		if (--iter <= 0) {
+			TRACE_ERROR("interrupt parsing due to max iteration reached",
+			            QUIC_EV_CONN_PRSHPKT, qc);
+			goto err;
+		}
+		else if (parsing_stage <= 1) {
+			TRACE_ERROR("interrupt parsing due to buffering blocked on gap size limit",
+			            QUIC_EV_CONN_PRSHPKT, qc);
 			goto err;
-
-		case QUIC_RX_RET_FRM_AGAIN:
-			TRACE_STATE("AGAIN encountered", QUIC_EV_CONN_PRSHPKT, qc);
-			cf_err = 1;
-			break;
-
-		case QUIC_RX_RET_FRM_DONE:
-			TRACE_PROTO("frame handled", QUIC_EV_CONN_PRSAFRM, qc, frm);
-			break;
-
-		case QUIC_RX_RET_FRM_DUP:
-			if (qc_is_listener(qc) && qel == qc->iel &&
-				!(qc->flags & QUIC_FL_CONN_HANDSHAKE_SPEED_UP)) {
-				fast_retrans = 1;
-			}
-			break;
 		}
 
-		eb64_delete(&frm->crypto.offset_node);
-		qc_frm_free(qc, &frm);
+		parsing_stage = 0;
+		list_for_each_entry_safe(frm, frm_tmp, &retry_frms, list) {
+			/* only CRYPTO frames may be reparsed for now */
+			BUG_ON(frm->type != QUIC_FT_CRYPTO);
+			ret = qc_handle_crypto_frm(qc, &frm->crypto, pkt, qel);
+			switch (ret) {
+			case QUIC_RX_RET_FRM_FATAL:
+				goto err;
+
+			case QUIC_RX_RET_FRM_AGAIN:
+				if (parsing_stage == 0) {
+					TRACE_STATE("parsing stage set to 1 (AGAIN encountered)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				break;
+
+			case QUIC_RX_RET_FRM_DONE:
+				TRACE_PROTO("frame handled after a new parsing iteration",
+				            QUIC_EV_CONN_PRSAFRM, qc, frm);
+				if (parsing_stage == 1) {
+					TRACE_STATE("parsing stage set to 2 (DONE after AGAIN)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				__fallthrough;
+			case QUIC_RX_RET_FRM_DUP:
+				qc_frm_free(qc, &frm);
+				break;
+			}
+		}
+
+		/* Always reset <frm> as it may be dangling after
+		 * list_for_each_entry_safe() usage. Especially necessary to
+		 * prevent a crash if loop is interrupted on max iteration.
+		 */
+		frm = NULL;
 	}
 
-	if (cf_err)
-		goto err;
-
 	/* Error should be returned if some frames cannot be parsed. */
-	BUG_ON(!eb_is_empty(&cf_frms_tree));
+	BUG_ON(!LIST_ISEMPTY(&retry_frms));
 
 	if (fast_retrans && qc->iel && qc->hel) {
 		struct quic_enc_level *iqel = qc->iel;
@@ -1158,11 +1185,7 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
  err:
 	if (frm)
 		qc_frm_free(qc, &frm);
-	node = eb64_first(&cf_frms_tree);
-	while (node) {
-		frm = eb64_entry(node, struct quic_frame, crypto.offset_node);
-		node = eb64_next(node);
-		eb64_delete(&frm->crypto.offset_node);
+	list_for_each_entry_safe(frm, frm_tmp, &retry_frms, list) {
 		qc_frm_free(qc, &frm);
 	}
 
@@ -1525,7 +1548,7 @@ static int qc_try_rm_hp(struct quic_conn *qc, struct quic_rx_packet *pkt,
 
 	ret = 1;
  out:
-	TRACE_LEAVE(QUIC_EV_CONN_TRMHP, qc, pkt);
+	TRACE_LEAVE(QUIC_EV_CONN_TRMHP, qc);
 	return ret;
 }
 
@@ -1627,7 +1650,7 @@ static inline int quic_padding_check(const unsigned char *pos,
 
 /* Validate the token, retry or not (provided by NEW_TOKEN) parsed into
  * <pkt> RX packet from <dgram> datagram.
- * Return 1 if succeded, 0 if not.
+ * Return 1 if succeeded, 0 if not.
  */
 static inline int quic_token_validate(struct quic_rx_packet *pkt,
                                       struct quic_dgram *dgram,
@@ -1785,19 +1808,15 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 			pkt->saddr = dgram->saddr;
 			ipv4 = dgram->saddr.ss_family == AF_INET;
 
-			conn_id = quic_cid_alloc(NULL);
-			if (!conn_id) {
-				TRACE_ERROR("error on first CID allocation",
-				            QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
+			/* Generate the first connection CID. This is derived from the client
+			 * ODCID and address. This allows to retrieve the connection from the
+			 * ODCID without storing it in the CID tree. This is an interesting
+			 * optimization as the client is expected to stop using its ODCID in
+			 * favor of our generated value.
+			 */
+			conn_id = new_quic_cid(NULL, NULL, &pkt->dcid, &pkt->saddr);
+			if (!conn_id)
 				goto err;
-			}
-
-			if (quic_cid_derive_from_odcid(conn_id, &pkt->dcid, &pkt->saddr)) {
-				TRACE_ERROR("error on CID generation",
-				            QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
-				pool_free(pool_head_quic_connection_id, conn_id);
-				goto err;
-			}
 
 			qc = qc_new_conn(pkt->version, ipv4, &pkt->dcid, &pkt->scid, &token_odcid,
 			                 conn_id, &dgram->daddr, &pkt->saddr, 1,
@@ -1807,26 +1826,24 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 				goto err;
 			}
 
-			conn_id->qc = qc;
 			/* Compute and store into the quic_conn the hash used to compute extra CIDs */
-			if (quic_hash64_from_cid) {
+			if (quic_hash64_from_cid)
 				qc->hash64 = quic_hash64_from_cid(conn_id->cid.data, conn_id->cid.len,
-				                                  global.cluster_secret, sizeof(global.cluster_secret));
-			}
+								  global.cluster_secret, sizeof(global.cluster_secret));
 
 			if (quic_cid_insert(conn_id, new_tid)) {
-				/* Collision during CID insertion. This may
-				 * happen when handling two Initial packets
-				 * from the same client on different threads.
-				 * <new_tid> will be set to redispatch the
-				 * current packet.
-				 */
 				pool_free(pool_head_quic_connection_id, conn_id);
 				quic_conn_release(qc);
 				qc = NULL;
 			}
 			else {
-				quic_cid_register_seq_num(conn_id);
+				/* From here, <qc> is the correct connection for this <pkt> Initial
+				 * packet. <conn_id> must be inserted in the CIDs tree for this
+				 * connection.
+				 */
+				eb64_insert(qc->cids, &conn_id->seq_num);
+				/* Initialize the next CID sequence number to be used for this connection. */
+				qc->next_cid_seq_num = 1;
 
 				if (dgram->flags & QUIC_DGRAM_FL_REJECT)
 					quic_set_connection_close(qc, quic_err_transport(QC_ERR_CONNECTION_REFUSED));
