@@ -11,6 +11,14 @@
 #include <haproxy/trace.h>
 
 DECLARE_POOL(pool_head_quic_ssl_sock_ctx, "quic_ssl_sock_ctx", sizeof(struct ssl_sock_ctx));
+const char *quic_ciphers = "TLS_AES_128_GCM_SHA256:TLS_AES_256_GCM_SHA384"
+                           ":TLS_CHACHA20_POLY1305_SHA256:TLS_AES_128_CCM_SHA256";
+#ifdef HAVE_OPENSSL_QUIC
+const char *quic_groups = "X25519:P-256:P-384:P-521:X25519MLKEM768";
+#else
+const char *quic_groups = "X25519:P-256:P-384:P-521";
+#endif
+
 
 /* Set the encoded version of the transport parameter into the TLS
  * stack depending on <ver> QUIC version and <server> boolean which must
@@ -558,7 +566,6 @@ static int ha_quic_ossl_got_transport_params(SSL *ssl, const unsigned char *para
 {
 	int ret = 0;
 	struct quic_conn *qc = SSL_get_ex_data(ssl, ssl_qc_app_data_index);
-	struct listener *l = objt_listener(qc->target);
 
 	TRACE_ENTER(QUIC_EV_TRANSP_PARAMS, qc);
 
@@ -567,7 +574,7 @@ static int ha_quic_ossl_got_transport_params(SSL *ssl, const unsigned char *para
 		            QUIC_EV_TRANSP_PARAMS, qc);
 		ret = 1;
 	}
-	else if (!quic_transport_params_store(qc, !l, params, params + params_len)) {
+	else if (!quic_transport_params_store(qc, 0, params, params + params_len)) {
 		goto err;
 	}
 
@@ -718,6 +725,26 @@ int ssl_quic_initial_ctx(struct bind_conf *bind_conf)
 	SSL_CTX_set_mode(ctx, SSL_MODE_RELEASE_BUFFERS);
 	SSL_CTX_set_min_proto_version(ctx, TLS1_3_VERSION);
 	SSL_CTX_set_max_proto_version(ctx, TLS1_3_VERSION);
+	if (SSL_CTX_set_ciphersuites(ctx, quic_ciphers) != 1) {
+		ha_warning("Binding [%s:%d] for %s %s: default QUIC cipher"
+		           " suites setting failed.\n",
+		           bind_conf->file, bind_conf->line,
+		           proxy_type_str(bind_conf->frontend),
+		           bind_conf->frontend->id);
+		cfgerr++;
+	}
+
+#ifndef HAVE_OPENSSL_QUICTLS
+	/* TODO: this should also work with QUICTLS */
+	if (SSL_CTX_set1_groups_list(ctx, quic_groups) != 1) {
+		ha_warning("Binding [%s:%d] for %s %s: default QUIC cipher"
+		           " groups setting failed.\n",
+		           bind_conf->file, bind_conf->line,
+		           proxy_type_str(bind_conf->frontend),
+		           bind_conf->frontend->id);
+		cfgerr++;
+	}
+#endif
 
 	if (bind_conf->ssl_conf.early_data) {
 #if !defined(HAVE_SSL_0RTT_QUIC)
@@ -1023,6 +1050,12 @@ int qc_ssl_provide_all_quic_data(struct quic_conn *qc, struct ssl_sock_ctx *ctx)
 /* Simple helper to set the specifig OpenSSL/quictls QUIC API callbacks */
 int quic_ssl_set_tls_cbs(SSL *ssl)
 {
+	struct quic_conn *qc = SSL_get_ex_data(ssl, ssl_qc_app_data_index);
+
+	/* Ignore the TCP connections */
+	if (!qc)
+		return 1;
+
 #ifdef HAVE_OPENSSL_QUIC
 	return SSL_set_quic_tls_cbs(ssl, ha_quic_dispatch, NULL);
 #else
